@@ -3,6 +3,7 @@ from foxes.core import PointDataModel, PointDataModelList
 import foxes.algorithms.downwind.models as dm
 import foxes.models as fm
 import foxes.variables as FV
+import foxes.constants as FC
 
 
 class Downwind(Algorithm):
@@ -41,11 +42,34 @@ class Downwind(Algorithm):
         e.g. `{"state": 1000}` for chunks of 1000 states
     dbook : foxes.DataBook, optional
         The data book, or None for default
+    keep_models : list of str
+        Keep these models data in memory and do not finalize them
     verbosity : int
         The verbosity level, 0 means silent
 
+    Attributes
+    ----------
+    states : foxes.core.States
+        The ambient states
+    wake_models : list of foxes.core.WakeModel
+        The wake models, applied to all turbines
+    rotor_model : foxes.core.RotorModel
+        The rotor model, for all turbines
+    wake_frame : foxes.core.WakeFrame
+        The wake frame
+    partial_wakes_model : foxes.core.PartialWakesModel
+        The partial wakes model
+    farm_controller : foxes.core.FarmController
+        The farm controller
+    n_states : int
+        The number of states
+
     """
 
+    FarmWakesCalculation = dm.FarmWakesCalculation
+    PointWakesCalculation = dm.point_wakes_calc.PointWakesCalculation
+    SetAmbPointResults = dm.set_amb_point_results.SetAmbPointResults
+    
     def __init__(
         self,
         mbook,
@@ -58,9 +82,10 @@ class Downwind(Algorithm):
         farm_controller="basic_ctrl",
         chunks={FV.STATE: 1000},
         dbook=None,
+        keep_models=[],
         verbosity=1,
     ):
-        super().__init__(mbook, farm, chunks, verbosity, dbook)
+        super().__init__(mbook, farm, chunks, verbosity, dbook, keep_models)
 
         self.states = states
         self.n_states = None
@@ -108,7 +133,6 @@ class Downwind(Algorithm):
             self.print(f"    {i}) {w}")
         self.print(deco)
         self.print(f"  turbine models:")
-        self.farm_controller.collect_models(self)
         for i, m in enumerate(self.farm_controller.pre_rotor_models.models):
             self.print(f"    {i}) {m} [pre-rotor]")
         for i, m in enumerate(self.farm_controller.post_rotor_models.models):
@@ -116,53 +140,107 @@ class Downwind(Algorithm):
                 f"    {i+len(self.farm_controller.pre_rotor_models.models)}) {m}"
             )
         self.print(deco)
-
-    def initialize(self, **states_init_pars):
+        self.print()
+    
+    def init_states(self):
         """
-        Initializes the algorithm.
-
-        Parameters
-        ----------
-        states_init_pars : dict, optional
-            Parameters for states initialization
-
+        Initialize states, if needed.
         """
         if not self.states.initialized:
-            self.print(f"\nInitializing states '{self.states.name}'")
-            self.states.initialize(self, verbosity=self.verbosity, **states_init_pars)
+            self.update_idata(self.states)
+            self.n_states = self.states.size()
 
-        self.n_states = self.states.size()
-        self.states_data = self.get_models_data(self.states)
-        self.print("States data:\n")
-        self.print(self.states_data)
-        super().initialize()
-
-    def reset_states(self, states, **states_init_pars):
+    def initialize(self):
         """
-        Reset the underlying states
-
-        Parameters
-        ----------
-        states : foxes.core.States
-            The new states
-
+        Initializes the algorithm.          
         """
-        if states is not self.states:
-            if self.initialized:
-                self.finalize(clear_mem=True)
-            self.states = states
-            self.initialize(**states_init_pars)
+        self.print(f"\nInitializing algorithm '{self.name}'")
+        super().initialize()           
+
+        self.init_states()
+
+        mdls = [
+            self.rotor_model,
+            self.farm_controller,
+            self.wake_frame,
+            self.partial_wakes_model,
+        ] + self.wake_models
+
+        self.update_idata(mdls)
+
+    def _collect_farm_models(
+            self,
+            vars_to_amb,
+            calc_parameters,
+            ambient,
+        ):
+        """
+        Helper function that creates model list
+        """
+        # prepare:
+        calc_pars = []
+        t2f = fm.farm_models.Turbine2FarmModel
+        mlist = FarmDataModelList(models=[])
+        mlist.name = f"{self.name}_calc"
+
+        # 0) set XHYD:
+        m = fm.turbine_models.SetXYHD()
+        m.name = "set_xyhd_tm"
+        mlist.models.append(t2f(m))
+        mlist.models[-1].name = "set_xyhd"
+        calc_pars.append(calc_parameters.get(mlist.models[-1].name, {}))
+
+        # 1) run pre-rotor turbine models via farm controller:
+        mlist.models.append(self.farm_controller)
+        calc_pars.append(calc_parameters.get(mlist.models[-1].name, {}))
+        calc_pars[-1]["pre_rotor"] = True
+
+        # 2) calculate yaw from wind direction at rotor centre:
+        mlist.models.append(fm.rotor_models.CentreRotor(calc_vars=[FV.WD, FV.YAW]))
+        mlist.models[-1].name = "calc_yaw"
+        calc_pars.append(calc_parameters.get(mlist.models[-1].name, {}))
+
+        # 3) calculate ambient rotor results:
+        mlist.models.append(self.rotor_model)
+        calc_pars.append(calc_parameters.get(mlist.models[-1].name, {}))
+        calc_pars[-1].update(
+            {"store_rpoints": True, "store_rweights": True, "store_amb_res": True}
+        )
+
+        # 4) calculate turbine order:
+        mlist.models.append(dm.CalcOrder())
+        mlist.models[-1].name = "calc_order"
+        calc_pars.append(calc_parameters.get(mlist.models[-1].name, {}))
+
+        # 5) run post-rotor turbine models via farm controller:
+        mlist.models.append(self.farm_controller)
+        calc_pars.append(calc_parameters.get(mlist.models[-1].name, {}))
+        calc_pars[-1]["pre_rotor"] = False
+
+        # 6) copy results to ambient, requires self.farm_vars:
+        self.farm_vars = mlist.output_farm_vars(self)
+        mlist.models.append(dm.SetAmbFarmResults(vars_to_amb))
+        mlist.models[-1].name = "set_amb_results"
+        calc_pars.append(calc_parameters.get(mlist.models[-1].name, {}))
+
+        # 7) calculate wake effects:
+        if not ambient:
+            mlist.models.append(self.FarmWakesCalculation())
+            mlist.models[-1].name = "calc_wakes"
+            calc_pars.append(calc_parameters.get(mlist.models[-1].name, {}))
+
+        # update variables:
+        self.farm_vars = [FV.WEIGHT] + mlist.output_farm_vars(self)
+
+        return mlist, calc_pars
 
     def calc_farm(
         self,
         vars_to_amb=None,
-        init_parameters={},
         calc_parameters={},
-        final_parameters={},
         persist=True,
-        clear_mem_models=True,
+        finalize=True,
         ambient=False,
-        **states_init_pars,
     ):
         """
         Calculate farm data.
@@ -172,25 +250,16 @@ class Downwind(Algorithm):
         vars_to_amb : list of str, optional
             Variables for which ambient variables should
             be stored. None means all.
-        init_parameters : dict
-            Parameters for model initialization.
-            Key: model name str, value: parameter dict
         calc_parameters : dict
             Parameters for model calculation.
             Key: model name str, value: parameter dict
-        final_parameters : dict
-            Parameters for model finalization.
-            Key: model name str, value: parameter dict
         persist : bool
             Switch for forcing dask to load all model data
             into memory
-        clear_mem_models : bool
-            Switch for clearing model memory during model
-            finalization
+        finalize : bool
+            Flag for finalization after calculation
         ambient : bool
             Flag for ambient instead of waked calculation
-        states_init_pars : dict, optional
-            Parameters for states initialization
 
         Returns
         -------
@@ -199,93 +268,23 @@ class Downwind(Algorithm):
             dimensions (state, turbine)
 
         """
-
+        # initialize algorithm:
         if not self.initialized:
-            self.initialize(**states_init_pars)
+            self.initialize()
 
         # welcome:
         self._print_deco("calc_farm")
 
-        # prepare:
-        init_pars = []
-        calc_pars = []
-        final_pars = []
-        t2f = fm.farm_models.Turbine2FarmModel
-        mlist = FarmDataModelList(models=[])
-        fdict = {"clear_mem": clear_mem_models}
+        # collect models:
+        mlist, calc_pars = self._collect_farm_models(
+            vars_to_amb, calc_parameters, ambient)
 
-        # 0) set XHYD:
-        mlist.models.append(t2f(fm.turbine_models.SetXYHD()))
-        mlist.models[-1].name = "set_xyhd"
-        init_pars.append(init_parameters.get(mlist.models[-1].name, {}))
-        calc_pars.append(calc_parameters.get(mlist.models[-1].name, {}))
-        final_pars.append(final_parameters.get(mlist.models[-1].name, fdict))
-
-        # 1) run pre-rotor turbine models via farm controller:
-        mlist.models.append(self.farm_controller)
-        init_pars.append(init_parameters.get(mlist.models[-1].name, {}))
-        calc_pars.append(calc_parameters.get(mlist.models[-1].name, {}))
-        final_pars.append(final_parameters.get(mlist.models[-1].name, fdict))
-        calc_pars[-1]["pre_rotor"] = True
-
-        # 2) calculate yaw from wind direction at rotor centre:
-        mlist.models.append(fm.rotor_models.CentreRotor(calc_vars=[FV.WD, FV.YAW]))
-        mlist.models[-1].name = "calc_yaw"
-        init_pars.append(init_parameters.get(mlist.models[-1].name, {}))
-        calc_pars.append(calc_parameters.get(mlist.models[-1].name, {}))
-        final_pars.append(final_parameters.get(mlist.models[-1].name, fdict))
-
-        # 3) calculate ambient rotor results:
-        mlist.models.append(self.rotor_model)
-        init_pars.append(init_parameters.get(mlist.models[-1].name, {}))
-        calc_pars.append(calc_parameters.get(mlist.models[-1].name, {}))
-        final_pars.append(final_parameters.get(mlist.models[-1].name, fdict))
-        calc_pars[-1].update(
-            {"store_rpoints": True, "store_rweights": True, "store_amb_res": True}
-        )
-
-        # 4) calculate turbine order:
-        mlist.models.append(dm.CalcOrder())
-        init_pars.append(init_parameters.get(mlist.models[-1].name, {}))
-        calc_pars.append(calc_parameters.get(mlist.models[-1].name, {}))
-        final_pars.append(final_parameters.get(mlist.models[-1].name, fdict))
-
-        # 5) run post-rotor turbine models via farm controller:
-        mlist.models.append(self.farm_controller)
-        init_pars.append(init_parameters.get(mlist.models[-1].name, {}))
-        calc_pars.append(calc_parameters.get(mlist.models[-1].name, {}))
-        final_pars.append(final_parameters.get(mlist.models[-1].name, fdict))
-        calc_pars[-1]["pre_rotor"] = False
-
-        # 6) copy results to ambient, requires self.farm_vars:
-        self.farm_vars = mlist.output_farm_vars(self)
-        mlist.models.append(dm.SetAmbFarmResults(vars_to_amb))
-        mlist.models[-1].name = "set_amb_results"
-        init_pars.append(init_parameters.get(mlist.models[-1].name, {}))
-        calc_pars.append(calc_parameters.get(mlist.models[-1].name, {}))
-        final_pars.append(final_parameters.get(mlist.models[-1].name, fdict))
-
-        # 7) calculate wake effects:
-        if not ambient:
-            mlist.models.append(dm.FarmWakesCalculation())
-            mlist.models[-1].name = "calc_wakes"
-            init_pars.append(init_parameters.get(mlist.models[-1].name, {}))
-            calc_pars.append(calc_parameters.get(mlist.models[-1].name, {}))
-            final_pars.append(final_parameters.get(mlist.models[-1].name, fdict))
-
-        # update variables:
-        self.farm_vars = [FV.WEIGHT] + mlist.output_farm_vars(self)
-
-        # initialize models:
-        mlist.initialize(self, parameters=init_pars, verbosity=self.verbosity)
-
-        # get input model data:
-        models_data = self.get_models_data(mlist).merge(
-            self.states_data, compat="identical"
-        )
+        # initialize models and get input model data:
+        self.update_idata(mlist)
+        models_data = self.get_models_data()
         if persist:
             models_data = models_data.persist()
-        self.print("\nInput model data:\n\n", models_data, "\n")
+        self.print("\nInput data:\n\n", models_data, "\n")
         self.print(f"\nOutput farm variables:", ", ".join(self.farm_vars))
         self.print(f"\nChunks: {self.chunks}\n")
 
@@ -296,13 +295,16 @@ class Downwind(Algorithm):
         farm_results = mlist.run_calculation(
             self, models_data, out_vars=self.farm_vars, parameters=calc_pars
         )
+        farm_results[FV.TNAME] = ((FV.TURBINE,), self.farm.turbine_names)
+        if FV.ORDER in farm_results:
+            farm_results[FV.ORDER] = farm_results[FV.ORDER].astype(FC.ITYPE)
         del models_data
 
         # finalize models:
-        self.print("\n")
-        mlist.finalize(
-            self, results=farm_results, parameters=final_pars, verbosity=self.verbosity
-        )
+        if finalize:
+            self.print("\n")
+            self.finalize_model(mlist)
+            self.finalize()
 
         if ambient:
             dvars = [v for v in farm_results.data_vars.keys() if v in FV.var2amb]
@@ -310,6 +312,63 @@ class Downwind(Algorithm):
 
         return farm_results
 
+    def _collect_point_models(
+            self,
+            vars,
+            vars_to_amb,
+            calc_parameters,
+            point_models,
+            ambient,
+        ):
+        """
+        Helper function that creates model list
+        """
+        # prepare:
+        calc_pars = []
+        mlist = PointDataModelList(models=[])
+
+        # prepare extra eval models:
+        emodels = []
+        emodels_cpars = []
+        if point_models is not None:
+            if not isinstance(point_models, list):
+                point_models = [point_models]
+            for m in point_models:
+                if isinstance(m, str):
+                    pname = m
+                    pmodel = self.mbook.point_models[pname]
+                    pmodel.name = pname
+                    emodels.append(pmodel)
+                elif isinstance(m, PointDataModel):
+                    emodels.append(m)
+                else:
+                    raise TypeError(f"Model '{m}' is neither str nor PointDataModel")
+                emodels_cpars.append(calc_parameters.get(emodels[-1].name, {}))
+        emodels = PointDataModelList(models=emodels)
+
+        # 0) calculate states results:
+        mlist.models.append(self.states)
+        calc_pars.append(calc_parameters.get(mlist.models[-1].name, {}))
+
+        # 1) calculate ambient extra eval point results:
+        mlist.models.append(emodels)
+        calc_pars.append({"parameters": emodels_cpars})
+
+        # 2) transfer ambient results:
+        mlist.models.append(
+            dm.SetAmbPointResults(point_vars=vars, vars_to_amb=vars_to_amb)
+        )
+        mlist.models[-1].name = "set_amb_results"
+        calc_pars.append(calc_parameters.get(mlist.models[-1].name, {}))
+
+        # 3) calc wake effects:
+        if not ambient:
+            mlist.models.append(dm.PointWakesCalculation(vars, emodels, emodels_cpars))
+            mlist.models[-1].name = "calc_wakes"
+            calc_pars.append(calc_parameters.get(mlist.models[-1].name, {}))
+
+        return mlist, calc_pars
+
     def calc_points(
         self,
         farm_results,
@@ -317,14 +376,11 @@ class Downwind(Algorithm):
         vars=None,
         vars_to_amb=None,
         point_models=None,
-        init_parameters={},
         calc_parameters={},
-        final_parameters={},
         persist_mdata=True,
         persist_pdata=False,
-        clear_mem_models=True,
+        finalize=True,
         ambient=False,
-        **states_init_pars,
     ):
         """
         Calculate data at a given set of points.
@@ -344,28 +400,19 @@ class Downwind(Algorithm):
             be stored. None means all.
         point_models : str or foxes.core.PointDataModel
             Additional point models to be executed
-        init_parameters : dict
-            Parameters for model initialization.
-            Key: model name str, value: parameter dict
         calc_parameters : dict
             Parameters for model calculation.
             Key: model name str, value: parameter dict
-        final_parameters : dict
-            Parameters for model finalization.
-            Key: model name str, value: parameter dict
         persist_mdata : bool
             Switch for forcing dask to load all model data
             into memory
         persist_fdata : bool
             Switch for forcing dask to load all farm data
             into memory
-        clear_mem_models : bool
-            Switch for clearing model memory during model
-            finalization
+        finalize : bool
+            Flag for finalization after calculation
         ambient : bool
             Flag for ambient instead of waked calculation
-        states_init_pars : dict, optional
-            Parameters for states initialization
 
         Returns
         -------
@@ -376,83 +423,27 @@ class Downwind(Algorithm):
         """
 
         if not self.initialized:
-            self.initialize(**states_init_pars)
+            self.initialize()
         if not ambient and farm_results is None:
             raise ValueError(
                 f"Cannot calculate point results without farm results for ambient = {ambient}"
             )
 
+        # welcome:
         self._print_deco("calc_points", n_points=points.shape[1])
 
-        # prepare:
-        init_pars = []
-        calc_pars = []
-        final_pars = []
-        mlist = PointDataModelList(models=[])
-        fdict = {"clear_mem": clear_mem_models}
+        # collect models:
+        mlist, calc_pars = self._collect_point_models(
+            vars, vars_to_amb, calc_parameters, point_models, ambient)
 
-        # prepare extra eval models:
-        emodels = []
-        emodels_ipars = []
-        emodels_cpars = []
-        emodels_fpars = []
-        if point_models is not None:
-            if not isinstance(point_models, list):
-                point_models = [point_models]
-            for m in point_models:
-                if isinstance(m, str):
-                    pname = m
-                    pmodel = self.mbook.point_models[pname]
-                    pmodel.name = pname
-                    emodels.append(pmodel)
-                elif isinstance(m, PointDataModel):
-                    emodels.append(m)
-                else:
-                    raise TypeError(f"Model '{m}' is neither str nor PointDataModel")
-                emodels_ipars.append(init_parameters.get(emodels[-1].name, {}))
-                emodels_cpars.append(calc_parameters.get(emodels[-1].name, {}))
-                emodels_fpars.append(final_parameters.get(emodels[-1].name, fdict))
-        emodels = PointDataModelList(models=emodels)
-
-        # 0) calculate states results:
-        mlist.models.append(self.states)
-        init_pars.append(init_parameters.get(mlist.models[-1].name, {}))
-        calc_pars.append(calc_parameters.get(mlist.models[-1].name, {}))
-        final_pars.append(final_parameters.get(mlist.models[-1].name, {}))
-
-        # 1) calculate ambient extra eval point results:
-        mlist.models.append(emodels)
-        init_pars.append({"parameters": emodels_ipars})
-        calc_pars.append({"parameters": emodels_cpars})
-        final_pars.append({"parameters": emodels_fpars})
-
-        # 2) transfer ambient results:
-        mlist.models.append(
-            dm.SetAmbPointResults(point_vars=vars, vars_to_amb=vars_to_amb)
-        )
-        mlist.models[-1].name = "set_amb_results"
-        init_pars.append(init_parameters.get(mlist.models[-1].name, {}))
-        calc_pars.append(calc_parameters.get(mlist.models[-1].name, {}))
-        final_pars.append(final_parameters.get(mlist.models[-1].name, fdict))
-
-        # 3) calc wake effects:
-        if not ambient:
-            mlist.models.append(dm.PointWakesCalculation(vars, emodels, emodels_cpars))
-            mlist.models[-1].name = "calc_wakes"
-            init_pars.append(init_parameters.get(mlist.models[-1].name, {}))
-            calc_pars.append(calc_parameters.get(mlist.models[-1].name, {}))
-            final_pars.append(final_parameters.get(mlist.models[-1].name, fdict))
-
-        # initialize models:
-        mlist.initialize(self, parameters=init_pars, verbosity=self.verbosity)
-
-        # get input model data:
-        models_data = self.get_models_data(mlist).merge(
-            self.states_data, compat="identical"
-        )
+        # initialize models and get input model data:
+        self.update_idata(mlist)
+        models_data = self.get_models_data()
         if persist_mdata:
             models_data = models_data.persist()
-        self.print("\nInput model data:\n\n", models_data, "\n")
+        self.print("\nInput data:\n\n", models_data, "\n")
+        self.print(f"\nOutput farm variables:", ", ".join(self.farm_vars))
+        self.print(f"\nChunks: {self.chunks}\n")
 
         # chunk farm results:
         if self.chunks is not None:
@@ -496,10 +487,10 @@ class Downwind(Algorithm):
         del models_data, farm_results, point_data
 
         # finalize models:
-        self.print("\n")
-        mlist.finalize(
-            self, point_results, parameters=final_pars, verbosity=self.verbosity
-        )
+        if finalize:
+            self.print("\n")
+            mlist.finalize(self, self.verbosity)
+            self.finalize()
 
         if ambient:
             dvars = [v for v in point_results.data_vars.keys() if v in FV.var2amb]
@@ -514,10 +505,18 @@ class Downwind(Algorithm):
         Parameters
         ----------
         clear_mem : bool
-            Flag for deleting algorithm data and
-            resetting initialization flag
+            Clear idata memory, including keep_models entries
 
         """
-        if clear_mem:
-            self.states_data = None
-        super().finalize(clear_mem=clear_mem)
+        mdls = [
+            self.states,
+            self.rotor_model,
+            self.farm_controller,
+            self.wake_frame,
+            self.partial_wakes_model,
+        ] + self.wake_models
+        
+        for m in mdls:
+            self.finalize_model(m)
+
+        super().finalize(clear_mem)
