@@ -4,3 +4,7 @@ from .set_farm_vars import SetFarmVars
 from .thrust2ct import Thrust2Ct
 from .power_mask import PowerMask
 from .sector_management import SectorManagement
+from .table_factors import TableFactors
+from .yaw2yawm import YAW2YAWM
+from .yawm2yaw import YAWM2YAW
+from .calculator import Calculator
