@@ -1,3 +1,5 @@
 from .farm_opt_problem import FarmOptProblem
+from .farm_vars_problem import FarmVarsProblem
 from .farm_objective import FarmObjective
 from .farm_constraint import FarmConstraint
+from .pop_states import PopStates
