@@ -1,6 +1,7 @@
 FARM = "farm"
 STATE = "state"
 TURBINE = "turbine"
+TNAME = "tname"
 POINT = "point"
 POINTS = "points"
 RPOINT = "rotor_point"
@@ -9,6 +10,7 @@ RWEIGHTS = "rotor_weights"
 ORDER = "order"
 
 VARS = "vars"
+VALID = "valid"
 TMODELS = "tmodels"
 TMODEL_SELS = "tmodel_sels"
 
@@ -77,6 +79,7 @@ amb2var = {a: v for v, a in var2amb.items()}
 AMB_RPOINT_RESULTS = "amb_rpoint_res"
 
 K = "k"
+KB = "kb"
 KY = "ky"
 KZ = "kz"
 KTI = "kTI"
@@ -85,6 +88,7 @@ KTIZ = "kTIz"
 
 Z0 = "z0"
 MOL = "MOL"
+SHEAR = "shear"
 
 POP = "pop"
 
